// --- [ atdb081.c ] ----------------------------------------------------------
//
// tab = 3
// ----------------------------------------------------------------------------

#include <avr/interrupt.h>
#include <avr/io.h>
#include <avr/eeprom.h>
#include <inttypes.h>
#include <stdlib.h>
#include <avr/pgmspace.h>
#include "TWI_df_slave.h"
#include "at45db081.h"
#include "spi_hard.h"
#include "rs232.h"
#include "global.h"

uint16_t df_page_empty;							// Anzahl der freien Pages - nur zum Debuggen
uint16_t df_page_used;							// Anzahl der benutzten Pages - nur zum Debuggen
uint16_t df_file_size;							// Anzahl der Bytes in der Datei - nur zum Debuggen

uint8_t df_adr[3];								// zur Adressierung des DF, die Adresse in Bytes aufbereitet

// --- lokale Funktionen ------------------------------------------------------
void DF_make_adr(uint16_t page,  uint16_t byte_adr);

// ----------------------------------------------------------------------------
// Schreibt eine komplette Page aus einem der beiden internen Puffer
// (in zwei Schritten) an DF.
// Zuerst werden die Daten in einen der beide DF-Buffer geschickt,
// dann wird das Kommando gegeben, den DF-Buffer in den Speicher des DF
// zu transferieren.
// Mit dem Parameter idx wird mitgeteilt, in welchen der beiden internen Puffer die zu
// schreibenden Daten sich befinden
// Da idx 0oder 1 betrgt, wird er auch gleich zur Auswahl der beiden Puffer auf dem DF
// herangezogen: die Daten werden abwechselnd in den Zwischenpuffer 1 bzw. 2 geschrieben
void DF_write_page(uint8_t idx)
{
uint8_t i = 0;

DF_make_adr(df_next_page, 0);					// fr die Aktion die Speicheradresse ermitteln

/* ------------------------------------*/ // Daten in den Buffer

// Im letzten Datensatz unbenutzte Bytes mit dem Wert 255 auffllen
// lt. Dokumentation sollten unbenutzte Bytes mit 255 beschrieben werden
if (bytes_received < 255)						// wenn weniger als 256 Byte in der Page vorhanden sind
	{													// die freien Bytes auf 255 setzen
	for (i = (bytes_received + 1); i; i++) rx_buffer[idx][i] = 255;
	}

SET_PIN_CS;											// cs setzen
if (idx)
	{
	SPI_write(BUFFER_1_WRITE);					// Kommando via Buffer 1 schreiben
	}

else
	{
	SPI_write(BUFFER_2_WRITE);					// Kommando via Buffer 2 schreiben
	}
SPI_write(0);										// Adr[0] ist hier ohne Bedeutung
SPI_write(df_adr[1]);
SPI_write(df_adr[2]);

i = 0;												// die Daten schreiben
do	{ SPI_write(rx_buffer[idx][i++]);		// das Array mit 256 Byte
	} while (i);
														// den Header vorbereiten
df_header[FILE_ID_HIGH]     = (uint8_t) (file_id >> 8);	// File-ID
df_header[FILE_ID_LOW]      = (uint8_t)  file_id;
														
df_header[PAGE_SIZE_HIGH]   = 0; //(uint8_t) ((bytes_received >> 8) & 0x0F);
df_header[PAGE_SIZE_LOW]    = (uint8_t) bytes_received;	// Pagesize; 0	bedeutet, dass 1 Byte eingegangen ist !

df_header[PAGE_CNT_HIGH]    = (uint8_t) (file_page_cnt >> 8);// Pagecounter fr Datei
df_header[PAGE_CNT_LOW]     = (uint8_t)  file_page_cnt;

df_header[PAGE_TIME_HIGH]   =  t.time8[1]; 	// die interne Zeit
df_header[PAGE_TIME_LOW]    =  t.time8[0];
df_header[PAGE_SIZE_HIGH]  += (t.time8[2] << 4);	
	
i = 0;												// dann den Header mit 8 Byte schreiben
do	{ SPI_write(df_header[i++]);
	} while (i < 8);

CLEAR_PIN_CS;

/* ------------------------------------*/ // Daten vom Buffer in den Speicher

while (!(DF_device_ready()));					// warten, bis DF bereit

SET_PIN_CS;											// cs setzen
if (idx)
	{													// Kommando Buffer 1 to Memory
	SPI_write(B1_TO_MM_PAGE_PROG_WITH_ERASE);
	}

else
	{													// Kommando Buffer 2 to Memory
	SPI_write(B2_TO_MM_PAGE_PROG_WITH_ERASE);
	}
SPI_write(df_adr[0]);
SPI_write(df_adr[1]);
SPI_write(0);										// Adr[2] ist hier ohne Bedeutung

CLEAR_PIN_CS;

/* ------------------------------------*/ // die nchste Page lschen

#ifdef NO_OVERWRITE								// siehe global.h
session_pages++;									// Zhler fr die geschriebenen Pages 
														// innerhalb einer Sitzung, wenn
if (session_pages & (!(PAGE_MASK)))			// mehr als 4096 Pages geschrieben
	{													// dann keine Daten berschreiben
	LED1_ON;
	LED2_ON;											// alle LEDs aus
	LED3_ON;
	while(1)											// alle drei LEDs im Gleichtakt toggeln
		{
		LED_timeout = 7;
		while (LED_timeout);
		LED1_TOG;
		LED2_TOG;
		LED3_TOG;
		}
	}
#endif

file_page_cnt++;									// Zhler fr die Pages einer Datei
df_next_page++;									// df_next_page zeigt immer auf die
}														// folgende Page

/*
// ----------------------------------------------------------------------------
// Lscht eine Page als Vorbereitung vor dem Schreiben von Daten
// hier muss die Page-nr per Value bergeben werden
// dauert 30ms !!
void DF_erase_page(uint16_t page)
{
while (!(DF_device_ready()));					// warten, bis DF bereit
DF_make_adr(page, 0);							// die Adresse berechnen
									
SET_PIN_CS;											// cs setzen

SPI_write(PAGE_ERASE);							// Kommando schreiben
SPI_write(df_adr[0]);
SPI_write(df_adr[1]);
SPI_write(0);										// hier ist kein Wert erforderlich

CLEAR_PIN_CS;										// cs lschen
}
*/

// ----------------------------------------------------------------------------
// gibt TRUE zurck, wenn das Gert bereit ist

uint8_t DF_device_ready(void)
{
uint8_t status;

SET_PIN_CS;											// cs setzen

SPI_write(STATUS_REGISTER);					// Kommando schreiben
status = SPI_read();								// Antwort abholen

CLEAR_PIN_CS;										// cs lschen

// damit ein 0xFF (eine ausbleibende Antwort) nicht zu einem TRUE fhrt, weitere Bits auswerten
//return(status & 0x80);						// Bit.7 maskieren
status = status & 0b10111100;					// Achtung - Gerteabhngig/Speicherabhngig !!!
if (status == STATUS_READY) return(TRUE);	// Achtung - Gerteabhngig/Speicherabhngig !!!
														// ergibt sich aus 128 + chipspezifischen Bits
return(FALSE );
}

// ----------------------------------------------------------------------------
// schreibt die Device_ID nach df_buffer[]
void DF_read_device_id(void)
{
uint8_t i;

SET_PIN_CS;											// cs setzen

SPI_write(DEVICE_ID);							// Kommando schreiben

for (i = 0; i < 6; i++)
	{
	df_buffer[i] = SPI_read();					// ID abholen
	}

CLEAR_PIN_CS;										// cs lschen
}

// ----------------------------------------------------------------------------
// gibt TRUE zurck, wenn beim letzten Vergleich Buffer-Memory einen Fehler auftrat
uint8_t DF_compare_failed(void)
{
uint8_t status;

SET_PIN_CS;											// cs setzen

SPI_write(STATUS_REGISTER);					// Kommando schreiben
status = SPI_read();								// Antwort abholen

CLEAR_PIN_CS;										// cs lschen

return(status & 0x40);							// Bit.6 maskieren
}

// ----------------------------------------------------------------------------
// Durchsucht den gesamten DF, ermittelt, wieviele Pages frei / belegt sind,
// die hchste Datei-ID,
// die Pagenummer fr den nchsten zu speichernde Page.
// Dazu werden aus allen Pages die Bytes 256 bis 264 eingelesen
// Wenn in PAGE_ID_HIGH das Bit.7 gesetzt ist, dann ist diese Page gelscht worden - also leer

									// Index/Bedeutung der Bytes im 'Header' jeder Page
// PAGE_ID_HIGH		0		// laufende Filenr./ID, beginnend mit 1
// PAGE_ID_LOW			1
// PAGE_SIZE_HIGH		2		// Lower Nibble: Anzahl der Bytes in der Page + PAGE_SIZE_LOW
									// Upper Nibble; Bit.16 - 20 der Zeitzhlung
// PAGE_SIZE_LOW		3		// Anzahl der Bytes in der Page (low byte)
// PAGE_CNT_HIGH		4		// Fortlaufende Zhlung der Pages einer Datei 
// PAGE_CNT_LOW		5
// PAGE_TIME_HIGH		6		// Zeit seit dem Start des Loggers in Sekunden
// PAGE_TIME_LOW		7		// plus upper nibble von PAGE_SIZE_HIGH

void DF_check(void)
{
uint8_t start_found = FALSE;
uint16_t ii, cc;

df_page_empty = 0;								// nur zur Information
df_page_used  = 0;								// nur zur Information
df_next_page  = 0;								// nchste frei Page
file_id       = 0;								// nchste File-Id ("Dateiname")

LED1_ON;												// LED1/2 unterschiedlich Schalten
LED2_OFF;

DF_read_byte(0, PAGE_SIZE, 4);				// den Header der ersten Page lesen
if (!(df_buffer[FILE_ID_HIGH] & 0x80))		// wenn Bit.7 nicht gesetzt ist, dann ist die Page benutzt
	{													// dann deren File-ID bernehmen
	file_id = ((df_buffer[FILE_ID_HIGH] << 8) + df_buffer[FILE_ID_LOW]);
	}

for (ii = 0; ii < (PAGES_ON_CHIP - 1); ii++)
	{
	DF_read_byte(ii, 256, 4);					// 4 Byte nach df_buffer einlesen

	if (!(df_buffer[FILE_ID_HIGH] & 0x80))	// wenn Bit.7 nicht gesetzt ist, dann ist die Page benutzt
		{
		df_page_used++; 							// Zhler fr benutzte Pages incrementieren
														// die 16Bit-ID nach cc
		cc = ((df_buffer[FILE_ID_HIGH] << 8) + df_buffer[FILE_ID_LOW]);
														// die File-IDs 
														// Vergleich mit der letzten File-ID
		if (cc == file_id)						// wenn cc == file_id, dann ist das ein Page der selben Datei
			{
			continue;								// dann gibts nichts weiter zu tun
			}
		
		else if (cc == file_id + 1)			// dann ist es eine Page der nachfolgenden Datei
			{
			file_id = cc;							// die hhere File_id merken
			}
		else 											// die File_Ids folgen nicht in Einerschritten aufeinander
			{											// wird als Ende der letzten Datei interpretiert
			if (!(start_found))					// wenn noch keine leere Page gefunden
				{
				start_found = TRUE;				// damit nachfolgende leeren Seiten ignoriert werden				
				df_next_page = ii;				// dann ist das die nchste freie Page
				}			
			}
		}

	else 												// wenn die Page nicht benutzt ist
		{
		df_page_empty++;
		if (!(start_found))						// wenn noch keine leere Page gefunden
			{
			start_found = TRUE;					// damit nachfolgende leeren Seiten ignoriert werden
			df_next_page = ii;					// das ist diese die nchste freie Page
			}
		}

	if (LED_timeout == 0)
		{
		LED_timeout = 3;							// schnelles Blinken
		LED1_TOG;									// beide LEDs togglen
		LED2_TOG;
		}
	}
}

// ----------------------------------------------------------------------------
// liest cnt-Bytes ab der angegebenen Adresse aus dem DF
// und schreibt die Daten nach df_buffer[]

void DF_read_byte(uint16_t page, uint16_t byte_adr, uint8_t cnt)
{
uint8_t i = 0;

DF_make_adr(page, byte_adr);
SET_PIN_CS;											// cs setzen

SPI_write(CONT_READ_LOW);						// Kommando schreiben
SPI_write(df_adr[0]);
SPI_write(df_adr[1]);
SPI_write(df_adr[2]);

if (cnt == 255)
	{
	do	{ df_buffer[i++] = SPI_read(); } while (i);
	}
else
	{
	do	{ df_buffer[i++] = SPI_read(); } while (i <= cnt);
	}

CLEAR_PIN_CS;
}

// ----------------------------------------------------------------------------
// liest 256 Byte einer Page beginnend mit der angegebenen Byte-Adresse aus dem DF
// und schreibt die Daten nach df_buffer[]

void DF_read_page(uint16_t page, uint16_t byte_adr)
{
uint8_t i = 0;

DF_make_adr(page, byte_adr);					// die Adresse umrechnen

SET_PIN_CS;											// cs setzen

SPI_write(MAIN_MEMORY_PAGE_READ);			// Kommando schreiben
SPI_write(df_adr[0]);
SPI_write(df_adr[1]);
SPI_write(df_adr[2]);

SPI_write(0);										// 4 dont care Bytes
SPI_write(0);
SPI_write(0);
SPI_write(0);

do	{ df_buffer[i++] = SPI_read();			// Daten nach sf_buffer[] schreiben
	} while (i);

CLEAR_PIN_CS;
}

// ----------------------------------------------------------------------------
// bildet aus den Variablen page und byte
// die zur Adressierung im 264-Byte Modus notwendige Adressbytes
void DF_make_adr(uint16_t page, uint16_t byte_adr)
{
page = page << 1;									// Page um 1 Bit nach links schieben

df_adr[0]  = (uint8_t) (page >> 8);
df_adr[1]  = (uint8_t)  page;
df_adr[1] += (uint8_t) (byte_adr  >> 8);	// das obere Bit der Adresse addieren
df_adr[2]  = (uint8_t)  byte_adr;
}

// ----------------------------------------------------------------------------
// lscht den gesamten Speicher,
// wartet bis zum Abschluss der Aktion, dabei blinken die beiden
// LEDs schnell im Gegentakt
void DF_erase_chip(void)
{
SET_PIN_CS;											// cs setzen

SPI_write(0xC7);									// feste Sequenz schreiben
SPI_write(0x94);
SPI_write(0x80);
SPI_write(0x9A);

CLEAR_PIN_CS;

LED1_ON;												// LEDs fr Gegentakt vorbereiten
LED2_OFF;

LED_timeout = 7;									// warten bis zum Abschluss des Lschens
while (!(DF_device_ready()))
	{
	if (LED_timeout == 0)						// dabei die LED schnell im Gegentakt blinken lassen
		{
		LED_timeout = 7;
		LED1_TOG;
		LED2_TOG;
		}
	}
}


// ----------------------------------------------------------------------------
// Druckt alle Pages einer Datei aus.
// bergeben wird die letzte Page der Datei,
// zurckgegeben wird die letzte Page der vorausgehenden Datei
// oder 0xFFFF bei einem Fehler.
uint16_t DF_print_file(uint16_t lastpage)
{
uint16_t startpage, endpage, page;			
uint16_t pages_in_file, file_id, ii;	
uint8_t i;						

endpage = lastpage;								// letzte Page speichern zum Debuggen
// die letzte Aufnahme wird durch ihre ID charakterisiert.
// es muss also solange zurckgespult werden, wie die ID identisch ist.
// zustzlich ist die Anzahl der Pages ja kodiert und kann zur Berechnung des
// Anfanges herangezogen werden, wenn man die Kette von hinten her aufdrselt
	
// Start dieser Datei ermitteln, dazu die Anzahl der Pages holen
// den Header einlesen
DF_read_byte(lastpage, PAGE_SIZE, 8);

// Jetzt steht der Header des letzen Datensatzes in df_buffer[]
pages_in_file  = (df_buffer[PAGE_CNT_HIGH] << 8) + df_buffer[PAGE_CNT_LOW];

file_id  = (df_buffer[FILE_ID_HIGH] << 8) + df_buffer[FILE_ID_LOW];
	
// hier sollte der Anfang der Datei sein
startpage = lastpage -(pages_in_file - 1) ;// Filesize beginnt mit 1						
startpage &= PAGE_MASK;						// verhindert Underflow

// Dorthin springen und den Header einlesen,
// wenn dort dieselbe fileid vorliegt und
// wenn filesize dort 1 ist, dann stimmt das Ziel.
DF_read_byte(startpage, PAGE_SIZE, 8);

// die FileId (low_Byte) vergleichen, wenn sie nicht bereinstimmt,
// dann haben wir die Datei verlassen.
if (!(file_id == ((df_buffer[FILE_ID_HIGH] << 8) + df_buffer[FILE_ID_LOW])))
	{
	put_c(13);
	puts_p(PSTR("end of files"));
	put_c(13);
	for (i = 0; i < 15; i++)
		{
		LED_timeout = 3;
		while (LED_timeout);
		LED2_TOG;
		}
	return(0xFFFF);
	}

// jetzt prfen, ob als filesize tatschlich 1 eingetragen ist.
// wenn nein, dann liegt ein Fehler vor. 	
if (!((df_buffer[PAGE_CNT_HIGH] == 0) && (df_buffer[PAGE_CNT_LOW] == 1)))
	{
	put_c(13);
	puts_p(PSTR("err: !cnt == 1"));
	put_c(13);
	for (i = 0; i < 15; i++)
		{
		LED_timeout = 3;
		while (LED_timeout);
		LED2_TOG;
		}
	return(0xFFFF);	
	}
// wenn der Eintrag fr filesize wie erwartet == 1 ist, dann die Daten ausgeben
else
	{
// die startpage kopieren, fr den Fall, das weitere Dateien gelesen werden sollen
	page = startpage;							// startpage kopieren nach page zum incrementieren
	df_file_size = 0;
// Pageweise die Daten rausschreiben
	for (ii = 0; ii < pages_in_file; ii++)
		{
		DF_print_page(page);					// die Page ausgeben
	
		page++;
		page &= PAGE_MASK;					// gegen Overflow am Ende des Speichers
	
		LED2_TOG;								// LED2 toggled nach jeweils nach einer Page
		}
	}

put_c(13);										// am Ende der Datei ein paar Infos ausgeben
put_c(13);										// kann alles entfallen !!!
puts_p(PSTR("file_ID:   "));
put_s(ultoa(file_id, msg, 10));
put_c(13);
puts_p(PSTR("startpage: "));
put_s(ultoa(startpage, msg, 10));
put_c(13);
puts_p(PSTR("endpage:   "));
put_s(ultoa(endpage, msg, 10));
put_c(13);
puts_p(PSTR("#pages:    "));
put_s(ultoa(pages_in_file, msg, 10));
put_c(13);
puts_p(PSTR("#bytes:    "));
put_s(ultoa(df_file_size, msg, 10));
put_c(13);

// Fr die Zeit noch einmal den Header der letzten Page einlesen
DF_read_byte(endpage, PAGE_SIZE, 8);

t.time8[3] = 0;
t.time8[2] = (df_buffer[PAGE_SIZE_HIGH] >> 4);// den Zeiteintrag der letzten Page
t.time8[1] = df_buffer[PAGE_TIME_HIGH];	// in die Variable time32 schreiben
t.time8[0] = df_buffer[PAGE_TIME_LOW];

puts_p(PSTR("seconds:   "));
put_s(ultoa(t.time32, msg, 10));
put_c(13);
put_c(13);

DF_check();											// die folgenden Werte neu ermitteln
	
puts_p(PSTR("nextpage:  "));
put_s(ultoa(df_next_page, msg, 10));
put_c(13);	
puts_p(PSTR("empty:     "));
put_s(ultoa(df_page_empty, msg, 10));
put_c(13);
puts_p(PSTR("used:      "));
put_s(ultoa(df_page_used, msg, 10));	
put_c(13);

startpage--;										// die Endpage der vorausgehenden
startpage &= PAGE_MASK;							// Datei berechnen
return(startpage);
}

// ----------------------------------------------------------------------------
// gibt eine Page von 256 Byte via RS232 aus
void DF_print_page(uint16_t page)
{
uint8_t cnt, i;

DF_read_byte(page, PAGE_SIZE, 4);			// den Header der Page einlesen
cnt = df_buffer[PAGE_SIZE_LOW];				// hier maximal 1 Byte
DF_read_byte(page, 0, 255);					// die Daten nach df_buffer[] einlesen
i = 0;		
if (cnt == 255)
	{
	do	{ put_c(df_buffer[i++]); } while (i);
	df_file_size += 256;
	}

else
	{
	do { put_c(df_buffer[i++]); } while (i <= cnt);
	df_file_size += (cnt + 1);
	}
}

// ----------------------------------------------------------------------------
// druckt alle Pages einschlielich des Headers zur Kontrolle aus.
// Endet, sobald das Bit.7 in PAGE_ID_HIGH gesetzt ist (gelschte Page).
void DF_file_test(void)
{
union {uint16_t uint16; uint8_t uint8[2];} x;
uint16_t page;

for (page = 0; page < PAGES_ON_CHIP; page++)
	{

	// zuerst den Header drucken
	DF_read_byte(page, PAGE_SIZE, 8);		// den Header nach df_buffer[] einlesen

	put_c(13);
	put_c(13);
	puts_p(PSTR("Page#: "));
	put_s(ultoa(page, msg, 10));
	put_c(',');

	if (df_buffer[FILE_ID_HIGH] & 0x80)		// testen, ob das Ende der Daten erreicht ist
		{
		puts_p(PSTR(" empty page."));
		put_c(13);
		return;
		}
		
	x.uint8[1] = df_buffer[FILE_ID_HIGH];
	x.uint8[0] = df_buffer[FILE_ID_LOW];
	puts_p(PSTR(" FileID: "));
	put_s(ultoa(x.uint16, msg, 10));
	put_c(',');

	x.uint8[1] = (df_buffer[PAGE_SIZE_HIGH] & 0x0F);
	x.uint8[0] = df_buffer[PAGE_SIZE_LOW];
	puts_p(PSTR(" Bytes: "));
	put_s(ultoa((x.uint16 + 1), msg, 10));
	put_c(',');

	x.uint8[1] = df_buffer[PAGE_CNT_HIGH];
	x.uint8[0] = df_buffer[PAGE_CNT_LOW];
	puts_p(PSTR(" cnt: "));
	put_s(ultoa(x.uint16 , msg, 10));
	put_c(',');
	t.time8[3] = 0;
	t.time8[2] = (df_buffer[PAGE_SIZE_HIGH] >> 4);
	t.time8[1] = df_buffer[PAGE_TIME_HIGH];
	t.time8[0] = df_buffer[PAGE_TIME_LOW];

	puts_p(PSTR(" time: "));
	put_s(ultoa(t.time32, msg, 10));

	put_c(13);
	put_c(13);

	DF_print_page(page);
	}
}

// --- [ eof ] ----------------------------------------------------------------